from pathlib import Path
from typing import Literal

import awkward as ak
import numba as nb
import numpy as np

from ...typing import BoolLike, FloatLike, IntLike

_cur_dir = Path(__file__).resolve().parent

_mdc_wire_position: dict[str, np.ndarray] = dict(np.load(_cur_dir / "mdc_geom.npz"))
_superlayer: np.ndarray = _mdc_wire_position["superlayer"]
_layer: np.ndarray = _mdc_wire_position["layer"]
_wire: np.ndarray = _mdc_wire_position["wire"]
_east_x: np.ndarray = _mdc_wire_position["east_x"]
_east_y: np.ndarray = _mdc_wire_position["east_y"]
_east_z: np.ndarray = _mdc_wire_position["east_z"]
_west_x: np.ndarray = _mdc_wire_position["west_x"]
_west_y: np.ndarray = _mdc_wire_position["west_y"]
_west_z: np.ndarray = _mdc_wire_position["west_z"]
_stereo: np.ndarray = _mdc_wire_position["stereo"]
_is_stereo: np.ndarray = _mdc_wire_position["is_stereo"]

# Generate the wire start index of each layer
layer_start_gid = np.zeros(44, dtype=np.uint16)
for l in range(43):
    layer_start_gid[l + 1] = np.sum(_layer == l)
layer_start_gid = np.cumsum(layer_start_gid)

# Generate the x position along z of each wire
dx_dz = (_east_x - _west_x) / (_east_z - _west_z)
dy_dz = (_east_y - _west_y) / (_east_z - _west_z)

# Generate layer -> is_stereo array
is_layer_stereo = np.zeros(43, dtype=bool)
for l in range(43):
    assert np.unique(_is_stereo[_layer == l]).size == 1
    is_layer_stereo[l] = _is_stereo[_layer == l][0]

# Generate layer -> superlayer array
superlayer_splits = np.array([0, 4, 8, 12, 16, 20, 24, 28, 32, 36, 40, 43])


def get_mdc_wire_position(library: Literal["np", "ak", "pd"] = "np"):
    """
    Get the MDC wire position table.

    Parameters:
        library: The library to return the data in. Choose from 'ak', 'np', 'pd'.

    Returns:
        (ak.Array | dict[str, np.ndarray] | pd.DataFrame): The MDC wire position table.

    Raises:
        ValueError: If the library is not 'ak', 'np', or 'pd'.
        ImportError: If the library is 'pd' but pandas is not installed.
    """
    cp: dict[str, np.ndarray] = {k: v.copy() for k, v in _mdc_wire_position.items()}

    if library == "ak":
        return ak.Array(cp)
    elif library == "np":
        return cp
    elif library == "pd":
        try:
            import pandas as pd  # type: ignore
        except ImportError:
            raise ImportError("Pandas is not installed. Run `pip install pandas`.")
        return pd.DataFrame(cp)
    else:
        raise ValueError(f"Invalid library {library}. Choose from 'ak', 'np', 'pd'.")


@nb.vectorize([nb.uint16(nb.int_, nb.int_)])
def get_mdc_gid(layer: IntLike, wire: IntLike) -> IntLike:
    """
    Get MDC gid of given layer and wire.

    Parameters:
        layer: The layer number.
        wire: The wire number.

    Returns:
        The gid of the wire.
    """
    return layer_start_gid[layer] + wire


@nb.vectorize([nb.uint8(nb.int_)])
def mdc_gid_to_superlayer(gid: IntLike) -> IntLike:
    """
    Convert gid to superlayer.

    Parameters:
        gid: The gid of the wire.

    Returns:
        The superlayer number of the wire.
    """
    return _superlayer[gid]


@nb.vectorize([nb.uint8(nb.int_)])
def mdc_layer_to_superlayer(layer: IntLike) -> IntLike:
    """
    Convert layer to superlayer.

    Parameters:
        layer: The layer number.

    Returns:
        The superlayer number of the layer.
    """
    return np.digitize(layer, superlayer_splits, right=True)


@nb.vectorize([nb.uint8(nb.int_)])
def mdc_gid_to_layer(gid: IntLike) -> IntLike:
    """
    Convert gid to layer.

    Parameters:
        gid: The gid of the wire.

    Returns:
        The layer number of the wire.
    """
    return _layer[gid]


@nb.vectorize([nb.uint16(nb.int_)])
def mdc_gid_to_wire(gid: IntLike) -> IntLike:
    """
    Convert gid to wire.

    Parameters:
        gid: The gid of the wire.

    Returns:
        The wire number of the wire.
    """
    return _wire[gid]


@nb.vectorize([nb.int8(nb.int_)])
def mdc_gid_to_stereo(gid: IntLike) -> IntLike:
    """
    Convert gid to stereo.
    `0` for `axial`,
    `-1` for stereo that `phi_west < phi_east`,
    `1` for stereo that `phi_west > phi_east`.

    Parameters:
        gid: The gid of the wire.

    Returns:
        The stereo of the wire.
    """
    return _stereo[gid]


@nb.vectorize([nb.boolean(nb.int_)])
def mdc_layer_to_is_stereo(layer: IntLike) -> BoolLike:
    """
    Convert layer to is_stereo.

    Parameters:
        layer: The layer number.

    Returns:
        The is_stereo of the layer.
    """
    return is_layer_stereo[layer]


@nb.vectorize([nb.boolean(nb.int_)])
def mdc_gid_to_is_stereo(gid: IntLike) -> BoolLike:
    """
    Convert gid to is_stereo.

    Parameters:
        gid: The gid of the wire.

    Returns:
        The is_stereo of the wire.
    """
    return _is_stereo[gid]


@nb.vectorize([nb.float64(nb.int_)])
def mdc_gid_to_west_x(gid: IntLike) -> FloatLike:
    """
    Convert gid to west_x (cm).

    Parameters:
        gid: The gid of the wire.

    Returns:
        The west_x (cm) of the wire.
    """
    return _west_x[gid]


@nb.vectorize([nb.float64(nb.int_)])
def mdc_gid_to_west_y(gid: IntLike) -> FloatLike:
    """
    Convert gid to west_y (cm).

    Parameters:
        gid: The gid of the wire.

    Returns:
        The west_y (cm) of the wire.
    """
    return _west_y[gid]


@nb.vectorize([nb.float64(nb.int_)])
def mdc_gid_to_west_z(gid: IntLike) -> FloatLike:
    """
    Convert gid to west_z (cm).

    Parameters:
        gid: The gid of the wire.

    Returns:
        The west_z (cm) of the wire.
    """
    return _west_z[gid]


@nb.vectorize([nb.float64(nb.int_)])
def mdc_gid_to_east_x(gid: IntLike) -> FloatLike:
    """
    Convert gid to east_x (cm).

    Parameters:
        gid: The gid of the wire.

    Returns:
        The east_x (cm) of the wire.
    """
    return _east_x[gid]


@nb.vectorize([nb.float64(nb.int_)])
def mdc_gid_to_east_y(gid: IntLike) -> FloatLike:
    """
    Convert gid to east_y (cm).

    Parameters:
        gid: The gid of the wire.

    Returns:
        The east_y (cm) of the wire.
    """
    return _east_y[gid]


@nb.vectorize([nb.float64(nb.int_)])
def mdc_gid_to_east_z(gid: IntLike) -> FloatLike:
    """
    Convert gid to east_z (cm).

    Parameters:
        gid: The gid of the wire.

    Returns:
        The east_z (cm) of the wire.
    """
    return _east_z[gid]


@nb.vectorize([nb.float64(nb.int_, nb.float64)])
def mdc_gid_z_to_x(gid: IntLike, z: FloatLike) -> FloatLike:
    """
    Get the x (cm) position of the wire at z (cm).

    Parameters:
        gid: The gid of the wire.
        z: The z (cm) position.

    Returns:
        The x (cm) position of the wire at z (cm).
    """
    return _west_x[gid] + dx_dz[gid] * (z - _west_z[gid])


@nb.vectorize([nb.float64(nb.int_, nb.float64)])
def mdc_gid_z_to_y(gid: IntLike, z: FloatLike) -> FloatLike:
    """
    Get the y (cm) position of the wire at z (cm).

    Parameters:
        gid: The gid of the wire.
        z: The z (cm) position.

    Returns:
        The y (cm) position of the wire at z (cm).
    """
    return _west_y[gid] + dy_dz[gid] * (z - _west_z[gid])



from pathlib import Path

import awkward as ak
import numpy as np

import pybes3 as p3
import pybes3.detectors as det

data_dir = Path(__file__).parent.parent / "data"
rtraw_event = p3.open(data_dir / "test_full_mc_evt_1.rtraw")["Event/TDigiEvent"].arrays()


def test_mdc_parse_gid():
    np_gid = det.get_mdc_wire_position()["gid"]
    ak_gid = ak.Array(np_gid)
    mdc_fields = [
        "gid",
        "layer",
        "wire",
        "stereo",
        "is_stereo",
        "superlayer",
        "mid_x",
        "mid_y",
        "west_x",
        "west_y",
        "west_z",
        "east_x",
        "east_y",
        "east_z",
    ]

    ak_res1 = p3.parse_mdc_gid(ak_gid, with_pos=True)
    assert ak_res1.fields == mdc_fields

    ak_res2 = p3.parse_mdc_gid(ak_gid, with_pos=False)
    assert ak_res2.fields == ["gid", "layer", "wire", "stereo", "is_stereo", "superlayer"]

    np_res1 = p3.parse_mdc_gid(np_gid, with_pos=True)
    assert list(np_res1.keys()) == mdc_fields

    np_res2 = p3.parse_mdc_gid(np_gid, with_pos=False)
    assert list(np_res2.keys()) == [
        "gid",
        "layer",
        "wire",
        "stereo",
        "is_stereo",
        "superlayer",
    ]


def test_parse_mdc_digi_id():
    mdc_id_ak: ak.Array = rtraw_event["m_mdcDigiCol"]["m_intId"]
    mdc_fields = [
        "gid",
        "layer",
        "wire",
        "stereo",
        "is_stereo",
        "superlayer",
        "mid_x",
        "mid_y",
        "west_x",
        "west_y",
        "west_z",
        "east_x",
        "east_y",
        "east_z",
    ]

    wire_ak = det.digi_id.mdc_id_to_wire(mdc_id_ak)
    layer_ak = det.digi_id.mdc_id_to_layer(mdc_id_ak)
    is_stereo_ak = det.digi_id.mdc_id_to_is_stereo(mdc_id_ak)

    # Test awkward, default option: with_pos=True
    ak_res1 = p3.parse_mdc_digi_id(mdc_id_ak, with_pos=True)
    assert ak_res1.fields == mdc_fields
    assert len(ak_res1.positional_axis) == 2
    assert ak.all(ak_res1["wire"] == wire_ak)
    assert ak.all(ak_res1["layer"] == layer_ak)
    assert ak.all(ak_res1["is_stereo"] == is_stereo_ak)

    mdc_id_np = ak.flatten(mdc_id_ak).to_numpy()
    wire_np = ak.flatten(wire_ak).to_numpy()
    layer_np = ak.flatten(layer_ak).to_numpy()
    is_stereo_np = ak.flatten(is_stereo_ak).to_numpy()

    # Test numpy, default option: with_pos=True
    np_res1 = p3.parse_mdc_digi_id(mdc_id_np, with_pos=True)
    assert list(np_res1.keys()) == mdc_fields
    assert np.all(np_res1["wire"] == wire_np)
    assert np.all(np_res1["layer"] == layer_np)
    assert np.all(np_res1["is_stereo"] == is_stereo_np)

    # Test int, default option: with_pos=True
    mdc_id_int = int(mdc_id_np[0])
    int_res1 = p3.parse_mdc_digi_id(mdc_id_int, with_pos=True)
    assert int_res1["wire"] == wire_np[0]
    assert int_res1["layer"] == layer_np[0]
    assert int_res1["is_stereo"] == is_stereo_np[0]

    # Test awkward, with_pos=False
    ak_res2 = p3.parse_mdc_digi_id(mdc_id_ak, with_pos=False)
    assert ak_res2.fields == ["gid", "layer", "wire", "stereo", "is_stereo", "superlayer"]
    assert len(ak_res2.positional_axis) == 2
    assert ak.all(ak_res2["wire"] == wire_ak)
    assert ak.all(ak_res2["layer"] == layer_ak)
    assert ak.all(ak_res2["is_stereo"] == is_stereo_ak)


def test_parse_mdc_digi():
    mdc_digi_ak: ak.Record = rtraw_event["m_mdcDigiCol"]
    base_fields = [
        "gid",
        "wire",
        "layer",
        "stereo",
        "is_stereo",
        "superlayer",
        "charge_channel",
        "time_channel",
        "track_index",
        "overflow",
        "digi_id",
    ]

    opt_fields = [
        "mid_x",
        "mid_y",
        "west_x",
        "west_y",
        "west_z",
        "east_x",
        "east_y",
        "east_z",
    ]

    ak_res1 = p3.parse_mdc_digi(mdc_digi_ak, with_pos=True)
    assert ak_res1.fields == base_fields + opt_fields
    assert len(ak_res1.positional_axis) == 2

    ak_res2 = p3.parse_mdc_digi(mdc_digi_ak, with_pos=False)
    assert ak_res2.fields == base_fields
    assert len(ak_res2.positional_axis) == 2


def test_parse_emc_gid():
    np_gid = det.get_emc_crystal_position()["gid"]
    ak_gid = ak.Array(np_gid)
    emc_fields = [
        "gid",
        "part",
        "theta",
        "phi",
        "front_center_x",
        "front_center_y",
        "front_center_z",
        "center_x",
        "center_y",
        "center_z",
    ]

    ak_res1 = p3.parse_emc_gid(ak_gid, with_pos=True)
    assert ak_res1.fields == emc_fields

    ak_res2 = p3.parse_emc_gid(ak_gid, with_pos=False)
    assert ak_res2.fields == ["gid", "part", "theta", "phi"]

    np_res1 = p3.parse_emc_gid(np_gid, with_pos=True)
    assert list(np_res1.keys()) == emc_fields

    np_res2 = p3.parse_emc_gid(np_gid, with_pos=False)
    assert list(np_res2.keys()) == ["gid", "part", "theta", "phi"]


def test_parse_emc_digi_id():
    emc_id_ak: ak.Array = rtraw_event["m_emcDigiCol"]["m_intId"]
    emc_fields = [
        "gid",
        "part",
        "theta",
        "phi",
        "front_center_x",
        "front_center_y",
        "front_center_z",
        "center_x",
        "center_y",
        "center_z",
    ]

    module_ak = det.digi_id.emc_id_to_module(emc_id_ak)
    theta_ak = det.digi_id.emc_id_to_theta(emc_id_ak)
    phi_ak = det.digi_id.emc_id_to_phi(emc_id_ak)

    # Test awkward, with_pos=True
    ak_res1 = p3.parse_emc_digi_id(emc_id_ak, with_pos=True)
    assert ak_res1.fields == emc_fields
    assert len(ak_res1.positional_axis) == 2
    assert ak.all(ak_res1["part"] == module_ak)
    assert ak.all(ak_res1["theta"] == theta_ak)
    assert ak.all(ak_res1["phi"] == phi_ak)

    emc_id_np = ak.flatten(emc_id_ak).to_numpy()
    module_np = ak.flatten(module_ak).to_numpy()
    theta_np = ak.flatten(theta_ak).to_numpy()
    phi_np = ak.flatten(phi_ak).to_numpy()

    # Test numpy, with_pos=True
    np_res1 = p3.parse_emc_digi_id(emc_id_np, with_pos=True)
    assert list(np_res1.keys()) == emc_fields
    assert np.all(np_res1["part"] == module_np)
    assert np.all(np_res1["theta"] == theta_np)
    assert np.all(np_res1["phi"] == phi_np)

    # Test int, with_pos=True
    emc_id_int = int(emc_id_np[0])
    int_res1 = p3.parse_emc_digi_id(emc_id_int, with_pos=True)
    assert int_res1["part"] == module_np[0]
    assert int_res1["theta"] == theta_np[0]
    assert int_res1["phi"] == phi_np[0]

    # Test awkward, with_pos=False
    ak_res2 = p3.parse_emc_digi_id(emc_id_ak, with_pos=False)
    assert ak_res2.fields == ["gid", "part", "theta", "phi"]
    assert len(ak_res2.positional_axis) == 2
    assert ak.all(ak_res2["part"] == module_ak)
    assert ak.all(ak_res2["theta"] == theta_ak)
    assert ak.all(ak_res2["phi"] == phi_ak)


def test_parse_emc_digi():
    emc_digi_ak: ak.Record = rtraw_event["m_emcDigiCol"]
    base_fields = [
        "gid",
        "part",
        "theta",
        "phi",
        "charge_channel",
        "time_channel",
        "track_index",
        "measure",
        "digi_id",
    ]

    opt_fields = [
        "front_center_x",
        "front_center_y",
        "front_center_z",
        "center_x",
        "center_y",
        "center_z",
    ]

    ak_res1 = p3.parse_emc_digi(emc_digi_ak, with_pos=True)
    assert ak_res1.fields == base_fields + opt_fields
    assert len(ak_res1.positional_axis) == 2

    ak_res2 = p3.parse_emc_digi(emc_digi_ak, with_pos=False)
    assert ak_res2.fields == base_fields
    assert len(ak_res2.positional_axis) == 2


def test_parse_tof_digi_id():
    tof_id_ak: ak.Array = p3.open(data_dir / "test_mrpc.rtraw")[
        "Event/TDigiEvent/m_tofDigiCol"
    ].array()["m_intId"]

    part_ak = det.digi_id.tof_id_to_part(tof_id_ak)
    layer_or_module_ak = det.digi_id.tof_id_to_layer_or_module(tof_id_ak)
    phi_or_strip_ak = det.digi_id.tof_id_to_phi_or_strip(tof_id_ak)
    end_ak = det.digi_id.tof_id_to_end(tof_id_ak)

    # Test awkward, flat=False, library='ak'
    ak_res1 = p3.parse_tof_digi_id(tof_id_ak, flat=False, library="ak")
    assert ak_res1.fields == ["part", "layer_or_module", "phi_or_strip", "end"]
    assert len(ak_res1.positional_axis) == 2
    assert ak.all(ak_res1["part"] == part_ak)
    assert ak.all(ak_res1["layer_or_module"] == layer_or_module_ak)
    assert ak.all(ak_res1["phi_or_strip"] == phi_or_strip_ak)
    assert ak.all(ak_res1["end"] == end_ak)

    # Test awkward, flat=True, library='ak'
    ak_res2 = p3.parse_tof_digi_id(tof_id_ak, flat=True, library="ak")
    assert ak_res2.fields == ["part", "layer_or_module", "phi_or_strip", "end"]
    assert len(ak_res2.positional_axis) == 1
    assert ak.all(ak_res2["part"] == ak.flatten(part_ak))
    assert ak.all(ak_res2["layer_or_module"] == ak.flatten(layer_or_module_ak))
    assert ak.all(ak_res2["phi_or_strip"] == ak.flatten(phi_or_strip_ak))
    assert ak.all(ak_res2["end"] == ak.flatten(end_ak))

    tof_id_np = ak.flatten(tof_id_ak).to_numpy()
    part_np = ak.flatten(part_ak).to_numpy()
    layer_or_module_np = ak.flatten(layer_or_module_ak).to_numpy()
    phi_or_strip_np = ak.flatten(phi_or_strip_ak).to_numpy()
    end_np = ak.flatten(end_ak).to_numpy()

    # Test numpy, library='np'
    np_res1 = p3.parse_tof_digi_id(tof_id_np, flat=False, library="np")
    assert list(np_res1.keys()) == ["part", "layer_or_module", "phi_or_strip", "end"]
    assert np.all(np_res1["part"] == part_np)
    assert np.all(np_res1["layer_or_module"] == layer_or_module_np)
    assert np.all(np_res1["phi_or_strip"] == phi_or_strip_np)
    assert np.all(np_res1["end"] == end_np)

    # Test int, library='ak'
    tof_id_int = int(tof_id_np[0])
    int_res1 = p3.parse_tof_digi_id(tof_id_int, flat=False, library="ak")
    assert int_res1.fields == ["part", "layer_or_module", "phi_or_strip", "end"]
    assert int_res1["part"] == part_np[0]
    assert int_res1["layer_or_module"] == layer_or_module_np[0]
    assert int_res1["phi_or_strip"] == phi_or_strip_np[0]
    assert int_res1["end"] == end_np[0]

    # Test int, library='np'
    int_res2 = p3.parse_tof_digi_id(tof_id_int, flat=False, library="np")
    assert list(int_res2.keys()) == ["part", "layer_or_module", "phi_or_strip", "end"]
    assert int_res2["part"] == part_np[0]
    assert int_res2["layer_or_module"] == layer_or_module_np[0]
    assert int_res2["phi_or_strip"] == phi_or_strip_np[0]
    assert int_res2["end"] == end_np[0]


def test_parse_muc_digi_id():
    muc_id_ak: ak.Array = p3.open(data_dir / "test_full_mc_evt_1.rtraw")[
        "Event/TDigiEvent/m_mucDigiCol"
    ].array()["m_intId"]

    part_ak = det.digi_id.muc_id_to_part(muc_id_ak)
    segment_ak = det.digi_id.muc_id_to_segment(muc_id_ak)
    layer_ak = det.digi_id.muc_id_to_layer(muc_id_ak)
    channel_ak = det.digi_id.muc_id_to_channel(muc_id_ak)
    gap_ak = det.digi_id.muc_id_to_gap(muc_id_ak)
    strip_ak = det.digi_id.muc_id_to_strip(muc_id_ak)

    # Test awkward, flat=False, library='ak'
    ak_res1 = p3.parse_muc_digi_id(muc_id_ak, flat=False, library="ak")
    assert ak_res1.fields == [
        "part",
        "segment",
        "layer",
        "channel",
        "gap",
        "strip",
    ]
    assert len(ak_res1.positional_axis) == 2
    assert ak.all(ak_res1["part"] == part_ak)
    assert ak.all(ak_res1["segment"] == segment_ak)
    assert ak.all(ak_res1["layer"] == layer_ak)
    assert ak.all(ak_res1["channel"] == channel_ak)
    assert ak.all(ak_res1["gap"] == gap_ak)
    assert ak.all(ak_res1["strip"] == strip_ak)

    # Test awkward, flat=True, library='ak'
    ak_res2 = p3.parse_muc_digi_id(muc_id_ak, flat=True, library="ak")
    assert ak_res2.fields == [
        "part",
        "segment",
        "layer",
        "channel",
        "gap",
        "strip",
    ]
    assert len(ak_res2.positional_axis) == 1
    assert ak.all(ak_res2["part"] == ak.flatten(part_ak))
    assert ak.all(ak_res2["segment"] == ak.flatten(segment_ak))
    assert ak.all(ak_res2["layer"] == ak.flatten(layer_ak))
    assert ak.all(ak_res2["channel"] == ak.flatten(channel_ak))
    assert ak.all(ak_res2["gap"] == ak.flatten(gap_ak))
    assert ak.all(ak_res2["strip"] == ak.flatten(strip_ak))

    muc_id_np = ak.flatten(muc_id_ak).to_numpy()
    part_np = ak.flatten(part_ak).to_numpy()
    segment_np = ak.flatten(segment_ak).to_numpy()
    layer_np = ak.flatten(layer_ak).to_numpy()
    channel_np = ak.flatten(channel_ak).to_numpy()
    gap_np = ak.flatten(gap_ak).to_numpy()
    strip_np = ak.flatten(strip_ak).to_numpy()

    # Test numpy, library='np'
    np_res1 = p3.parse_muc_digi_id(muc_id_np, flat=False, library="np")
    assert list(np_res1.keys()) == [
        "part",
        "segment",
        "layer",
        "channel",
        "gap",
        "strip",
    ]
    assert np.all(np_res1["part"] == part_np)
    assert np.all(np_res1["segment"] == segment_np)
    assert np.all(np_res1["layer"] == layer_np)
    assert np.all(np_res1["channel"] == channel_np)
    assert np.all(np_res1["gap"] == gap_np)
    assert np.all(np_res1["strip"] == strip_np)


def test_parse_cgem_digi_id():
    cgem_id_ak: ak.Array = p3.open(data_dir / "test_cgem.rtraw")[
        "Event/TDigiEvent/m_cgemDigiCol"
    ].array()["m_intId"]

    layer_ak = det.digi_id.cgem_id_to_layer(cgem_id_ak)
    sheet_ak = det.digi_id.cgem_id_to_sheet(cgem_id_ak)
    strip_ak = det.digi_id.cgem_id_to_strip(cgem_id_ak)
    is_x_strip_ak = det.digi_id.cgem_id_to_is_x_strip(cgem_id_ak)

    # Test awkward, flat=False, library='ak'
    ak_res1 = p3.parse_cgem_digi_id(cgem_id_ak, flat=False, library="ak")
    assert ak_res1.fields == ["layer", "sheet", "strip", "is_x_strip"]
    assert len(ak_res1.positional_axis) == 2
    assert ak.all(ak_res1["layer"] == layer_ak)
    assert ak.all(ak_res1["sheet"] == sheet_ak)
    assert ak.all(ak_res1["strip"] == strip_ak)
    assert ak.all(ak_res1["is_x_strip"] == is_x_strip_ak)

    # Test awkward, flat=True, library='ak'
    ak_res2 = p3.parse_cgem_digi_id(cgem_id_ak, flat=True, library="ak")
    assert ak_res2.fields == ["layer", "sheet", "strip", "is_x_strip"]
    assert len(ak_res2.positional_axis) == 1
    assert ak.all(ak_res2["layer"] == ak.flatten(layer_ak))
    assert ak.all(ak_res2["sheet"] == ak.flatten(sheet_ak))
    assert ak.all(ak_res2["strip"] == ak.flatten(strip_ak))
    assert ak.all(ak_res2["is_x_strip"] == ak.flatten(is_x_strip_ak))

    cgem_id_np = ak.flatten(cgem_id_ak).to_numpy()
    layer_np = ak.flatten(layer_ak).to_numpy()
    strip_np = ak.flatten(strip_ak).to_numpy()
    sheet_np = ak.flatten(sheet_ak).to_numpy()
    is_x_strip_np = ak.flatten(is_x_strip_ak).to_numpy()

    # Test numpy, library='np'
    np_res1 = p3.parse_cgem_digi_id(cgem_id_np, flat=False, library="np")
    assert list(np_res1.keys()) == ["layer", "sheet", "strip", "is_x_strip"]
    assert np.all(np_res1["layer"] == layer_np)
    assert np.all(np_res1["sheet"] == sheet_np)
    assert np.all(np_res1["strip"] == strip_np)
    assert np.all(np_res1["is_x_strip"] == is_x_strip_np)

    # Test int, library='ak'
    cgem_id_int = int(cgem_id_np[0])
    int_res1 = p3.parse_cgem_digi_id(cgem_id_int, flat=False, library="ak")
    assert int_res1.fields == ["layer", "sheet", "strip", "is_x_strip"]
    assert int_res1["layer"] == layer_np[0]
    assert int_res1["sheet"] == sheet_np[0]
    assert int_res1["strip"] == strip_np[0]
    assert int_res1["is_x_strip"] == is_x_strip_np[0]

    # Test int, library='np'
    int_res2 = p3.parse_cgem_digi_id(cgem_id_int, flat=False, library="np")
    assert list(int_res2.keys()) == ["layer", "sheet", "strip", "is_x_strip"]
    assert int_res2["layer"] == layer_np[0]
    assert int_res2["sheet"] == sheet_np[0]
    assert int_res2["strip"] == strip_np[0]
    assert int_res2["is_x_strip"] == is_x_strip_np[0]



from collections import deque
from typing import Literal, Union

import awkward as ak
import awkward.contents
import awkward.index
import numpy as np
import uproot
import uproot.behaviors
import uproot.behaviors.TBranch
import uproot.interpretation
import uproot.interpretation.identify
import uproot.interpretation.library
import uproot.models.TBranch

from .besio_cpp import read_bes_stl, read_bes_tobjarray, read_bes_tobject

##########################################################################################
#                                       Constants
##########################################################################################
kBASE = "BASE"
kCtype = "ctype"
kTString = "TString"
kVector = "vector"
kList = "list"
kDeque = "deque"
kMap = "map"
kSet = "set"
kMultiMap = "multimap"
kMultiSet = "multiset"

kTArrayC = "TArrayC"
kTArrayS = "TArrayS"
kTArrayI = "TArrayI"
kTArrayL = "TArrayL"
kTArrayF = "TArrayF"
kTArrayD = "TArrayD"

CTYPE_NAMES = {
    "bool",
    "char",
    "short",
    "int",
    "long",
    "float",
    "double",
    "unsigned char",
    "unsigned short",
    "unsigned int",
    "unsigned long",
    "int8_t",
    "uint8_t",
    "int16_t",
    "uint16_t",
    "int32_t",
    "uint32_t",
    "int64_t",
    "uint64_t",
}

STL_NAMES = {kVector, kList, kDeque, kMap, kSet, kMultiMap, kMultiSet}

TARRAY_NAMES = {kTArrayC, kTArrayS, kTArrayI, kTArrayL, kTArrayF, kTArrayD}

BES_BRANCH_NAMES = {
    "TEvtHeader",
    "TMcEvent",
    "TDigiEvent",
    "TDstEvent",
    "TRecEvent",
    "TEvtRecObject",
    "THltEvent",
}

kFallBack = "fallback"
kTObjArray = "TObjArray"
kTObject = "TObject"
kSTL = "STL"
branchname_to_element_type = {
    "TEvtHeader/m_eventId": (kFallBack, None),
    "TEvtHeader/m_runId": (kFallBack, None),
    "TEvtHeader/m_time": (kFallBack, None),
    "TEvtHeader/m_eventTag": (kFallBack, None),
    "TEvtHeader/m_flag1": (kFallBack, None),
    "TEvtHeader/m_flag2": (kFallBack, None),
    "TEvtHeader/m_etsT1": (kFallBack, None),
    "TEvtHeader/m_etsT2": (kFallBack, None),
    # multimap and map are the same
    "EventNavigator/m_mcMdcMcHits": (kSTL, "map<int,int>"),
    "EventNavigator/m_mcMdcTracks": (kSTL, "map<int,int>"),
    "EventNavigator/m_mcEmcMcHits": (kSTL, "map<int,int>"),
    "EventNavigator/m_mcEmcRecShowers": (kSTL, "map<int,int>"),
    "TMcEvent/m_mdcMcHitCol": (kTObjArray, "TMdcMc"),
    "TMcEvent/m_cgemMcHitCol": (kTObjArray, "TCgemMc"),
    "TMcEvent/m_emcMcHitCol": (kTObjArray, "TEmcMc"),
    "TMcEvent/m_tofMcHitCol": (kTObjArray, "TTofMc"),
    "TMcEvent/m_mucMcHitCol": (kTObjArray, "TMucMc"),
    "TMcEvent/m_mcParticleCol": (kTObjArray, "TMcParticle"),
    "TDigiEvent/m_fromMc": (kFallBack, None),
    "TDigiEvent/m_mdcDigiCol": (kTObjArray, "TMdcDigi"),
    "TDigiEvent/m_cgemDigiCol": (kTObjArray, "TCgemDigi"),
    "TDigiEvent/m_emcDigiCol": (kTObjArray, "TEmcDigi"),
    "TDigiEvent/m_tofDigiCol": (kTObjArray, "TTofDigi"),
    "TDigiEvent/m_mucDigiCol": (kTObjArray, "TMucDigi"),
    "TDigiEvent/m_lumiDigiCol": (kTObjArray, "TLumiDigi"),
    "TDstEvent/m_mdcTrackCol": (kTObjArray, "TMdcTrack"),
    "TDstEvent/m_emcTrackCol": (kTObjArray, "TEmcTrack"),
    "TDstEvent/m_tofTrackCol": (kTObjArray, "TTofTrack"),
    "TDstEvent/m_mucTrackCol": (kTObjArray, "TMucTrack"),
    "TDstEvent/m_mdcDedxCol": (kTObjArray, "TMdcDedx"),
    "TDstEvent/m_extTrackCol": (kTObjArray, "TExtTrack"),
    "TDstEvent/m_mdcKalTrackCol": (kTObjArray, "TMdcKalTrack"),
    "TRecEvent/m_recMdcTrackCol": (kTObjArray, "TRecMdcTrack"),
    "TRecEvent/m_recMdcHitCol": (kTObjArray, "TRecMdcHit"),
    "TRecEvent/m_recEmcHitCol": (kTObjArray, "TRecEmcHit"),
    "TRecEvent/m_recEmcClusterCol": (kTObjArray, "TRecEmcCluster"),
    "TRecEvent/m_recEmcShowerCol": (kTObjArray, "TRecEmcShower"),
    "TRecEvent/m_recTofTrackCol": (kTObjArray, "TRecTofTrack"),
    "TRecEvent/m_recMucTrackCol": (kTObjArray, "TRecMucTrack"),
    "TRecEvent/m_recMdcDedxCol": (kTObjArray, "TRecMdcDedx"),
    "TRecEvent/m_recMdcDedxHitCol": (kTObjArray, "TRecMdcDedxHit"),
    "TRecEvent/m_recExtTrackCol": (kTObjArray, "TRecExtTrack"),
    "TRecEvent/m_recMdcKalTrackCol": (kTObjArray, "TRecMdcKalTrack"),
    "TRecEvent/m_recMdcKalHelixSegCol": (kTObjArray, "TRecMdcKalHelixSeg"),
    "TRecEvent/m_recEvTimeCol": (kTObjArray, "TRecEvTime"),
    "TRecEvent/m_recZddChannelCol": (kTObjArray, "TRecZddChannel"),
    # This is an TObject, not a TObjArray
    "TEvtRecObject/m_evtRecEvent": (kTObject, "TEvtRecEvent"),
    "TEvtRecObject/m_evtRecTrackCol": (kTObjArray, "TEvtRecDTag"),
    # This is an TObject, not a TObjArray
    "TEvtRecObject/m_evtRecPrimaryVertex": (kTObject, "TEvtRecPrimaryVertex"),
    "TEvtRecObject/m_evtRecVeeVertexCol": (kTObjArray, "TEvtRecDTag"),
    "TEvtRecObject/m_evtRecPi0Col": (kTObjArray, "TEvtRecDTag"),
    "TEvtRecObject/m_evtRecEtaToGGCol": (kTObjArray, "TEvtRecDTag"),
    "TEvtRecObject/m_evtRecDTagCol": (kTObjArray, "TEvtRecDTag"),
    "THltEvent/m_hltRawCol": (kTObjArray, "THltRaw"),
    # This is an TObject, not a TObjArray
    "THltEvent/m_hltInf": (kTObject, "THltInf"),
    # This is an TObject, not a TObjArray
    "THltEvent/m_dstHltInf": (kTObject, "TDstHltInf"),
}


##########################################################################################
#                                 C++ output to awkward
##########################################################################################


def get_top_type_name(type_name: str) -> str:
    if type_name in CTYPE_NAMES:
        return kCtype

    elif "<" in type_name:
        top_type_name = type_name.split("<")[0]
        if top_type_name in STL_NAMES:
            return top_type_name
        else:
            raise NotImplementedError(f"Unsupported type_name: {type_name}")

    elif type_name == kTString:
        return kTString

    else:
        if type_name.endswith("*"):
            type_name = type_name[:-1]
        return type_name


def get_vector_subtype(type_name: str) -> str:
    assert type_name.startswith("vector<") and type_name.endswith(
        ">"
    ), f"Unsupported type_name: {type_name}"
    return type_name[7:-1].strip()


def get_map_subtypes(type_name: str) -> tuple[str, str]:
    assert type_name.startswith("map<") and type_name.endswith(
        ">"
    ), f"Unsupported type_name: {type_name}"

    pos_split = 0
    n_level = 0
    for i, c in enumerate(type_name):
        if c == "<":
            n_level += 1
        elif c == ">":
            n_level -= 1
        if n_level == 1 and c == ",":
            pos_split = i
            break

    assert pos_split != 0, f"Unsupported type_name: {type_name}"
    return type_name[4:pos_split].strip(), type_name[pos_split + 1 : -1].strip()


def recover_basic_element(
    streamer_info: dict, org_data: tuple, all_streamer_info: dict
) -> awkward.contents.Content:
    fTypeName = streamer_info["fTypeName"]
    fName = streamer_info["fName"]
    top_type_name = get_top_type_name(fTypeName)

    if "fArrayDim" in streamer_info and streamer_info["fArrayDim"] > 0:
        fMaxIndex = streamer_info["fMaxIndex"]

        new_info = streamer_info.copy()
        new_info["fArrayDim"] = 0

        flat_content = recover_basic_element(new_info, org_data, all_streamer_info)
        shape = deque(fMaxIndex[fMaxIndex > 0].tolist())
        return recover_array_shape(flat_content, shape)

    if top_type_name == kCtype:
        return awkward.contents.NumpyArray(org_data[0])

    elif top_type_name == kTString:
        offset, data = org_data
        return awkward.contents.ListOffsetArray(
            awkward.index.Index(offset),
            awkward.contents.NumpyArray(data.view(np.uint8), parameters={"__array__": "char"}),
            parameters={"__array__": "string"},
        )

    elif top_type_name in STL_NAMES:
        if top_type_name == kVector:
            element_type = get_vector_subtype(fTypeName)
            element_info = {
                "fName": fName + "_element",
                "fTypeName": element_type,
                "fType": -1,
            }

            offsets, sub_org_data = org_data
            sub_content = recover_basic_element(element_info, sub_org_data, all_streamer_info)

            return awkward.contents.ListOffsetArray(awkward.index.Index(offsets), sub_content)

        elif top_type_name == kMap or top_type_name == kMultiMap:
            key_type, val_type = get_map_subtypes(fTypeName)

            key_info = {"fName": fName + "_key", "fTypeName": key_type, "fType": -1}
            val_info = {"fName": fName + "_val", "fTypeName": val_type, "fType": -1}

            offset, key_org_data, val_org_data = org_data

            key_content = recover_basic_element(key_info, key_org_data, all_streamer_info)
            val_content = recover_basic_element(val_info, val_org_data, all_streamer_info)

            return awkward.contents.ListOffsetArray(
                awkward.index.Index(offset),
                awkward.contents.RecordArray([key_content, val_content], ["key", "val"]),
            )
        else:
            raise NotImplementedError(f"Unsupported STL type: {top_type_name}")

    elif top_type_name == kBASE:
        fType = streamer_info["fType"]
        if fType == 66:
            return None

        elif fType == 0:
            sub_streamers = all_streamer_info[fName]
            sub_field_names = []
            sub_field_contents = []
            for s_info, s_data in zip(sub_streamers, org_data):
                sub_content = recover_basic_element(s_info, s_data, all_streamer_info)
                sub_name = s_info["fName"]

                # skip TObject result
                if sub_content is not None:
                    sub_field_contents.append(sub_content)
                    sub_field_names.append(sub_name)
            return awkward.contents.RecordArray(sub_field_contents, sub_field_names)

        else:
            raise NotImplementedError(f"Unsupported fType: {fType}")

    elif top_type_name in TARRAY_NAMES:
        offsets, sub_org_data = org_data
        return awkward.contents.ListOffsetArray(
            awkward.index.Index(offsets), awkward.contents.NumpyArray(sub_org_data)
        )

    else:
        sub_streamers = all_streamer_info[top_type_name]
        sub_field_names = []
        sub_field_contents = []
        for s_info, s_data in zip(sub_streamers, org_data):
            sub_content = recover_basic_element(s_info, s_data, all_streamer_info)
            sub_name = s_info["fName"]

            # skip TObject result
            if sub_content is not None:
                sub_field_contents.append(sub_content)
                sub_field_names.append(sub_name)

        return awkward.contents.RecordArray(sub_field_contents, sub_field_names)


def recover_array_shape(
    content: awkward.contents.Content, shape: deque[int]
) -> awkward.contents.Content:
    cur_shape = shape.popleft()
    if len(shape) > 0:
        content = recover_array_shape(content, shape)
    return awkward.contents.RegularArray(content, cur_shape)


def tobjarray_np2ak(
    element_type_name: str, org_array_dict: dict, all_streamer_info: dict
) -> awkward.Array:
    org_data_tuple: tuple = org_array_dict["data"]
    obj_offsets: np.ndarray = org_array_dict["obj_offsets"]
    element_streamer_info = all_streamer_info[element_type_name]

    field_names = []
    field_contents = []
    for s, org_data in zip(element_streamer_info, org_data_tuple):
        element_content = recover_basic_element(s, org_data, all_streamer_info)
        if element_content is None:
            continue

        # Make object-offsets, then it's event-by-event array
        evt_contents = awkward.contents.ListOffsetArray(
            awkward.index.Index(obj_offsets), element_content
        )

        field_names.append(s["fName"])
        field_contents.append(evt_contents)

    return awkward.Array(awkward.contents.RecordArray(field_contents, field_names))


def tobject_np2ak(
    element_type_name: str, org_data_list: tuple, all_streamer_info: dict
) -> awkward.Array:
    field_names = []
    field_contents = []
    element_streamer_info = all_streamer_info[element_type_name]
    for s, org_data in zip(element_streamer_info, org_data_list):
        element_content = recover_basic_element(s, org_data, all_streamer_info)
        if element_content is None:
            continue

        field_names.append(s["fName"])
        field_contents.append(element_content)

    return awkward.Array(awkward.contents.RecordArray(field_contents, field_names))


def stl_np2ak(org_data: tuple, type_name: str, all_streamer_info: dict) -> awkward.Array:
    tmp_info = {
        "fName": "tmp",
        "fTypeName": type_name,
        "fType": -1,
    }
    element_content = recover_basic_element(tmp_info, org_data, all_streamer_info)
    return awkward.Array(element_content)


##########################################################################################
#                                     Array Preprocess
##########################################################################################
def get_symetric_matrix_idx(
    i: Union[int, ak.Array, np.ndarray], j: Union[int, ak.Array, np.ndarray], ndim: int
) -> int:
    r"""
    Returns the index of the similarity matrix given the row and column indices.

    The matrix is assumed to be symmetric-like. (i, j) -> index relationship is:

    |     | i=0 | i=1 | i=2 |
    | :-: | :-: | :-: | :-: |
    | j=0 |  0  |     |     |
    | j=1 |  1  |  2  |     |
    | j=2 |  3  |  4  |  5  |

    Parameters:
        i (Union[int, ak.Array, np.ndarray]): The row index or array of row indices.
        j (Union[int, ak.Array, np.ndarray]): The column index or array of column indices.
        ndim (int): The dimension of the similarity matrix.

    Returns:
        The index or array of indices corresponding to the given row and column indices.

    Raises:
        ValueError: If the row and column indices are not of the same type, or if one of them is not an integer.
        ValueError: If the row or column indices are greater than or equal to the dimension of the similarity matrix.
        ValueError: If the row or column indices are negative.
    """
    # Check type
    return_type: Literal["ak", "np"] = "ak"
    if type(i) != type(j):
        if isinstance(i, int):
            return_type = "np" if isinstance(j, np.ndarray) else "ak"
            i = ak.ones_like(j) * i
        elif isinstance(j, int):
            return_type = "np" if isinstance(i, np.ndarray) else "ak"
            j = ak.ones_like(i) * j
        else:
            raise ValueError(
                "i and j should be the same type, or one of them should be an integer."
            )
    else:
        return_type = "np" if isinstance(i, np.ndarray) else "ak"

    i, j = ak.sort([i, j], axis=0)
    res = j * (j + 1) // 2 + i

    # Check dimension
    if ak.any([i >= ndim, j >= ndim]):
        raise ValueError(
            "Indices i and j should be less than the dimension of the similarity matrix."
        )
    if ak.any([i < 0, j < 0]):
        raise ValueError("Indices i and j should be non-negative.")

    if return_type == "np" and isinstance(res, ak.Array):
        res = res.to_numpy()

    return res


def expand_zipped_symetric_matrix(
    arr: Union[ak.Array, np.ndarray],
) -> Union[ak.Array, np.ndarray]:
    """
    Recover a flattened simplified symmetric matrix represented as a 1D array back to a 2D matrix.
    This function assumes the last dimension of the input array is the flattened symmetric matrix,
    and will transform array

    ```
    [[a11, a12, a22, a13, a23, a33],
     [b11, b12, b22, b13, b23, b33]]
    ```

    to

    ```
    [[[a11, a12, a13],
      [a12, a22, a23],
      [a13, a23, a33]],

     [[b11, b12, b13],
      [b12, b22, b23],
      [b13, b23, b33]]]
    ```

    Parameters:
        arr (Union[ak.Array, np.ndarray]): The input array representing the flattened simplified symmetric matrix.

    Returns:
        The reshaped symmetric matrix as a 2D array.

    Raises:
        ValueError: If the input array does not have a symmetric shape.
    """
    ndim_data: int = arr.ndim

    # Get the number of elements in the symmetric matrix
    if isinstance(arr, ak.Array):
        type_strs = [i.strip() for i in arr.typestr.split("*")[:-1]]
        n_err_elements = int(type_strs[-1])

        ndim_err = (np.sqrt(1 + 8 * n_err_elements) - 1) / 2
        if not ndim_err.is_integer():
            raise ValueError("The array does not have a symmetric shape.")
        ndim_err = int(ndim_err)
    elif isinstance(arr, np.ndarray):
        n_err_elements = arr.shape[-1]

    # Reshape the array
    tmp_matrix = []
    for i in range(ndim_err):
        tmp_row = []
        for j in range(ndim_err):
            tmp_idx = tuple(
                [slice(None)] * (ndim_data - 1)
                + [get_symetric_matrix_idx(i, j, ndim_err), np.newaxis, np.newaxis]
            )
            tmp_row.append(arr[tmp_idx])

        if isinstance(arr, ak.Array):
            tmp_matrix.append(ak.concatenate(tmp_row, axis=-1))
        else:
            tmp_matrix.append(np.concatenate(tmp_row, axis=-1))

    if isinstance(arr, ak.Array):
        res = ak.concatenate(tmp_matrix, axis=-2)
    else:
        res = np.concatenate(tmp_matrix, axis=-2)

    return res


def expand_subbranch_symetric_matrix(
    sub_br_arr: ak.Array, matrix_fields: Union[str, list[str]]
) -> ak.Array:
    """
    Recover simplified symmetric matrix back to 2D matrix from specified fields of a branch array.

    Parameters:
        sub_br_arr: Subbranch array that need to be recovered.
        matrix_fields: Name of list of names of fields to be recovered.

    Returns:
        An array with recovered fields.
    """
    if isinstance(matrix_fields, str):
        matrix_fields = {matrix_fields}
    matrix_fields = set(matrix_fields)

    res_dict = {}
    for field_name in sub_br_arr.fields:
        if field_name in matrix_fields:
            res_dict[field_name] = expand_zipped_symetric_matrix(sub_br_arr[field_name])
        else:
            res_dict[field_name] = sub_br_arr[field_name]
    return ak.Array(res_dict)


#############################################
# TDigiEvent
#############################################
def process_digi_subbranch(org_arr: ak.Array) -> ak.Array:
    """
    Processes the `TRawData` subbranch of the input awkward array and returns a new array with the subbranch fields
    merged into the top level.

    Parameters:
        org_arr (ak.Array): The input awkward array containing the `TRawData` subbranch.

    Returns:
        A new awkward array with the fields of `TRawData` merged into the top level.

    Raises:
        AssertionError: If `TRawData` is not found in the input array fields.
    """
    assert "TRawData" in org_arr.fields, "TRawData not found in the input array"

    fields = {}
    for field_name in org_arr.fields:
        if field_name == "TRawData":
            for raw_field_name in org_arr[field_name].fields:
                fields[raw_field_name] = org_arr[field_name][raw_field_name]
        else:
            fields[field_name] = org_arr[field_name]

    return ak.Array(fields)


#############################################
# TEvtRecObject
#############################################
def process_evtrec_m_Evtx(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_Evtx")


def process_evtrec_m_evtRecVeeVertexCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_Ew")


#############################################
# TDstEvent
#############################################
def process_dst_m_mdcTrackCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_err")


def process_dst_m_emcTrackCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_err")


def process_dst_m_extTrackCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(
        org_arr,
        {
            "myEmcErrorMatrix",
            "myMucErrorMatrix",
            "myTof1ErrorMatrix",
            "myTof2ErrorMatrix",
        },
    )


#############################################
# TRecEvent
#############################################
def process_rec_m_recMdcTrackCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_err")


def process_rec_m_recEmcShowerCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_err")


def process_rec_m_recMdcKalTrackCol(org_arr: ak.Array) -> ak.Array:
    return expand_subbranch_symetric_matrix(org_arr, "m_terror")


#############################################
# Main function
#############################################
def preprocess_subbranch(full_branch_path: str, org_arr: ak.Array) -> ak.Array:
    evt_name, subbranch_name = full_branch_path.split("/")

    if evt_name == "TDigiEvent" and subbranch_name != "m_fromMc":
        return process_digi_subbranch(org_arr)

    if evt_name == "TEvtRecObject":
        if subbranch_name == "m_Evtx":
            return process_evtrec_m_Evtx(org_arr)
        if subbranch_name == "m_evtRecVeeVertexCol":
            return process_evtrec_m_evtRecVeeVertexCol(org_arr)

    if evt_name == "TDstEvent":
        if subbranch_name == "m_mdcTrackCol":
            return process_dst_m_mdcTrackCol(org_arr)
        if subbranch_name == "m_emcTrackCol":
            return process_dst_m_emcTrackCol(org_arr)
        if subbranch_name == "m_extTrackCol":
            return process_dst_m_extTrackCol(org_arr)

    if evt_name == "TRecEvent":
        if subbranch_name == "m_recMdcTrackCol":
            return process_rec_m_recMdcTrackCol(org_arr)
        if subbranch_name == "m_recEmcShowerCol":
            return process_rec_m_recEmcShowerCol(org_arr)
        if subbranch_name == "m_recMdcKalTrackCol":
            return process_rec_m_recMdcKalTrackCol(org_arr)

    # Default return
    return org_arr


##########################################################################################
#                                      Interpretation
##########################################################################################


class BesInterpretation(uproot.interpretation.Interpretation):
    def __init__(self, full_branch_name: str) -> None:
        super().__init__()

        self.full_branch_name = full_branch_name
        self.parse_rule, self.element_type_name = branchname_to_element_type[full_branch_name]

    def final_array(
        self,
        basket_arrays,
        entry_start,
        entry_stop,
        entry_offsets,
        library,
        branch,
        options,
    ):
        basket_entry_starts = np.array(entry_offsets[:-1])
        basket_entry_stops = np.array(entry_offsets[1:])

        basket_start_idx = np.where(basket_entry_starts <= entry_start)[0].max()
        basket_end_idx = np.where(basket_entry_stops >= entry_stop)[0].min()

        arr_to_concat = [basket_arrays[i] for i in range(basket_start_idx, basket_end_idx + 1)]
        tot_array = ak.concatenate(arr_to_concat)

        relative_entry_start = entry_start - basket_entry_starts[basket_start_idx]
        relative_entry_stop = entry_stop - basket_entry_starts[basket_start_idx]

        return tot_array[relative_entry_start:relative_entry_stop]

    def __eq__(self, other):
        return isinstance(other, self.__class__)

    def __repr__(self) -> str:
        return f"BES::As({self.typename})"

    @property
    def cache_key(self):
        return str(self.__class__)

    @property
    def typename(self):
        res = "BES::"
        if self.parse_rule == kTObjArray:
            res += f"TObjArray<{self.element_type_name}>"
        elif self.parse_rule == kTObject:
            res += self.element_type_name
        elif self.parse_rule == kSTL:
            res += self.element_type_name
        else:
            raise NotImplementedError(
                f"Unsupported parse_rule: {self.parse_rule}, {self.element_type_name}"
            )
        return res

    def basket_array(
        self,
        data: np.ndarray,
        byte_offsets: np.ndarray,
        basket,
        branch: uproot.models.TBranch.Model_TBranchElement,
        context,
        cursor_offset,
        library,
        options,
    ):
        # For BES TObject and TObjArray, element class streamers are needed
        # Check streamers and versions
        if self.parse_rule == kTObjArray or self.parse_rule == kTObject:
            obj_streamer = branch.file.streamers.get(self.element_type_name)
            if obj_streamer is None:
                raise ValueError(
                    f"Streamer for {self.element_type_name} not found, maybe this branch is empty?"
                )

            obj_versions = list(obj_streamer.keys())
            assert (
                len(obj_versions) == 1
            ), "Only one version of streamer is supported. Maybe you mixed data from different versions?"

        def get_streamer_list(type_name: str) -> list[dict]:
            return [
                i.all_members
                for i in list(branch.file.streamers[type_name].values())[0].member("fElements")
            ]

        all_streamer_info = {k: get_streamer_list(k) for k in branch.file.streamers.keys()}

        if self.parse_rule == kTObjArray:
            org_data = read_bes_tobjarray(
                data, byte_offsets, self.element_type_name, all_streamer_info
            )
        elif self.parse_rule == kTObject:
            org_data = read_bes_tobject(
                data, byte_offsets, self.element_type_name, all_streamer_info
            )
        elif self.parse_rule == kSTL:
            org_data = read_bes_stl(
                data, byte_offsets, self.element_type_name, all_streamer_info
            )
        else:
            raise NotImplementedError(
                f"Unsupported parse_rule: {self.parse_rule}, {self.element_type_name}"
            )

        if isinstance(library, uproot.interpretation.library.NumPy):
            raise NotImplementedError("Numpy library is not supported")

        if isinstance(library, uproot.interpretation.library.Pandas):
            raise NotImplementedError("Pandas library is not supported")

        assert isinstance(library, uproot.interpretation.library.Awkward), (
            "Unknown library: %s" % library
        )

        if self.parse_rule == kTObjArray:
            ak_arr = tobjarray_np2ak(self.element_type_name, org_data, all_streamer_info)
        elif self.parse_rule == kTObject:
            ak_arr = tobject_np2ak(self.element_type_name, org_data, all_streamer_info)
        elif self.parse_rule == kSTL:
            ak_arr = stl_np2ak(org_data, self.element_type_name, all_streamer_info)
        else:
            raise NotImplementedError(
                f"Unsupported parse_rule: {self.parse_rule}, {self.element_type_name}"
            )

        # preprocess awkward array
        return preprocess_subbranch(self.full_branch_name, ak_arr)


##########################################################################################
#                                       Wrappers
##########################################################################################
_is_TBranchElement_branches_wrapped = False
_is_uproot_interpretation_of_wrapped = False

_uproot_interpretation_of = uproot.interpretation.identify.interpretation_of


def bes_interpretation_of(
    branch: uproot.behaviors.TBranch.TBranch, context: dict, simplify: bool = True
) -> uproot.interpretation.Interpretation:
    if not hasattr(branch, "parent"):
        return _uproot_interpretation_of(branch, context, simplify)

    parent_name = branch.parent.name
    full_branch_name = f"{parent_name}/{branch.name}"

    if full_branch_name not in branchname_to_element_type.keys():
        return _uproot_interpretation_of(branch, context, simplify)

    if branchname_to_element_type[full_branch_name][0] == kFallBack:
        return _uproot_interpretation_of(branch, context, simplify)

    return BesInterpretation(full_branch_name=full_branch_name)


def wrap_uproot_interpretation():
    global _is_uproot_interpretation_of_wrapped
    if not _is_uproot_interpretation_of_wrapped:
        _is_uproot_interpretation_of_wrapped = True
        uproot.interpretation.identify.interpretation_of = bes_interpretation_of


def wrap_uproot_TBranchElement_branches():
    def branches(self):
        if self.name not in BES_BRANCH_NAMES:
            return self.member("fBranches")
        else:
            res = []
            for br in self.member("fBranches"):
                if br.name == "TObject":
                    continue

                interpret_type, class_name = branchname_to_element_type[
                    f"{self.name}/{br.name}"
                ]
                if interpret_type == kFallBack or class_name in self.file.streamers:
                    res.append(br)
                else:
                    continue
            return res

    global _is_TBranchElement_branches_wrapped
    if not _is_TBranchElement_branches_wrapped:
        _is_TBranchElement_branches_wrapped = True
        uproot.models.TBranch.Model_TBranchElement.branches = property(branches)
        for v in uproot.models.TBranch.Model_TBranchElement.known_versions.values():
            v.branches = property(branches)


def wrap_uproot():
    """
    Wraps the uproot functions to use the BES interpretation.
    """
    wrap_uproot_interpretation()
    wrap_uproot_TBranchElement_branches()



from typing import TypeVar

import awkward as ak
import numpy as np

IntLike = TypeVar("ak.Array | np.ndarray | int", ak.Array, np.ndarray, int)
FloatLike = TypeVar("ak.Array | np.ndarray | float", ak.Array, np.ndarray, float)
BoolLike = TypeVar("ak.Array | np.ndarray | bool", ak.Array, np.ndarray, bool)



from pathlib import Path

import awkward as ak

import pybes3 as p3

data_dir = Path(__file__).parent.parent / "data"

helix_ak = p3.open(data_dir / "test_full_mc_evt_1.dst")[
    "Event/TDstEvent/m_mdcTrackCol"
].array()["m_helix"]

helix_np = ak.flatten(helix_ak, axis=1).to_numpy()


def test_parse_helix():
    fields = [
        "x",
        "y",
        "z",
        "r",
        "px",
        "py",
        "pz",
        "pt",
        "p",
        "theta",
        "phi",
        "charge",
        "r_trk",
    ]

    p_helix_ak1 = p3.parse_helix(helix_ak)
    assert p_helix_ak1.fields == fields

    p_helix_ak2 = p3.parse_helix(helix_np, library="ak")
    assert p_helix_ak2.fields == fields

    p_helix_np = p3.parse_helix(helix_np)
    assert list(p_helix_np.keys()) == fields



from typing import Any

import uproot

from .raw_io import RawBinaryReader
from .raw_io import concatenate as concatenate_raw
from .root_io import wrap_uproot


def open(file, **kwargs) -> Any:
    """
    A wrapper around `uproot.open` that automatically calls `wrap_uproot` before opening the file.

    Parameters:
        file (str | Path | IO | dict[str | Path | IO, str]): The file to open.
        **kwargs (dict): Additional arguments to pass to `uproot.open`.

    Returns:
        The uproot file object.
    """
    wrap_uproot()
    return uproot.open(file, **kwargs)


def concatenate(files, branch: str, **kwargs) -> Any:
    """
    A wrapper around `uproot.concatenate` that automatically calls `wrap_uproot` before concatenating the files.

    Parameters:
        files (list[str | Path | IO, str]): The files to concatenate.
        branch (str): The branch to concatenate.
        **kwargs (dict): Additional arguments to pass to `uproot.concatenate`.

    Returns:
        The concatenated array.
    """
    wrap_uproot()
    return uproot.concatenate({str(f): branch for f in files}, **kwargs)


def open_raw(file: str) -> RawBinaryReader:
    """
    Open a raw binary file.

    Parameters:
        file (str): The file to open.

    Returns:
        (RawBinaryReader): The raw binary reader.
    """
    return RawBinaryReader(file)


__all__ = ["open", "concatenate", "open_raw", "concatenate_raw", "wrap_uproot"]



from typing import Literal, Union

import awkward as ak
import numba as nb
import numpy as np

from .typing import FloatLike


@nb.vectorize([nb.float64(nb.float64, nb.float64)])
def _helix01_to_x(helix0: FloatLike, helix1: FloatLike) -> FloatLike:
    """
    Convert helix parameters to x location.

    Parameters:
        helix0: helix[0] parameter, dr.
        helix1: helix[1] parameter, phi0.

    Returns:
        x location of the helix.
    """
    return helix0 * np.cos(helix1)


@nb.vectorize([nb.float64(nb.float64, nb.float64)])
def _helix01_to_y(helix0: FloatLike, helix1: FloatLike) -> FloatLike:
    """
    Convert helix parameters to y location.

    Parameters:
        helix0: helix[0] parameter, dr.
        helix1: helix[1] parameter, phi0.

    Returns:
        y location of the helix.
    """
    return helix0 * np.sin(helix1)


@nb.vectorize([nb.float64(nb.float64)])
def _helix2_to_pt(
    helix2: FloatLike,
) -> FloatLike:
    """
    Convert helix parameter to pt.

    Parameters:
        helix2: helix[2] parameter, kappa.

    Returns:
        pt of the helix.
    """
    return 1 / np.abs(helix2)


@nb.vectorize([nb.int8(nb.float64)])
def _helix2_to_charge(
    helix2: FloatLike,
) -> FloatLike:
    """
    Convert helix parameter to charge.

    Parameters:
        helix2: helix[2] parameter, kappa.

    Returns:
        charge of the helix.
    """
    if -1e-10 < helix2 < 1e-10:
        return 0
    return 1 if helix2 > 0 else -1


@nb.vectorize([nb.float64(nb.float64, nb.float64)])
def _pt_helix1_to_px(pt: FloatLike, helix1: FloatLike) -> FloatLike:
    """
    Convert pt and helix1 to px.

    Parameters:
        pt: pt of the helix.
        helix1: helix[1] parameter, phi0.

    Returns:
        px of the helix.
    """
    return -pt * np.sin(helix1)


@nb.vectorize([nb.float64(nb.float64, nb.float64)])
def _pt_helix1_to_py(pt: FloatLike, helix1: FloatLike) -> FloatLike:
    """
    Convert pt and helix1 to py.

    Parameters:
        pt: pt of the helix.
        helix1: helix[1] parameter, phi0.

    Returns:
        py of the helix.
    """
    return pt * np.cos(helix1)


@nb.vectorize([nb.float64(nb.float64, nb.float64)])
def _pt_helix4_to_p(pt: FloatLike, helix4: FloatLike) -> FloatLike:
    """
    Convert pt and helix4 to p.

    Parameters:
        pt: pt of the helix.
        helix4: helix[4] parameter, tanl.

    Returns:
        p of the helix.
    """
    return pt * np.sqrt(1 + helix4**2)


@nb.vectorize([nb.float64(nb.float64, nb.float64)])
def _pz_p_to_theta(pz: FloatLike, p: FloatLike) -> FloatLike:
    """
    Convert pz and p to theta.

    Parameters:
        pz: pz of the helix.
        p: p of the helix.

    Returns:
        theta of the helix.
    """
    return np.acos(pz / p)


def parse_helix(
    helix: Union[ak.Array, np.ndarray], library: Literal["ak", "auto"] = "auto"
) -> Union[ak.Array, dict[str, np.ndarray]]:
    """
    Parse helix parameters to physical parameters.

    Parameters:
        helix: helix parameters, the last dimension should be 5.
        library: the library to use, if "auto", return a dict when input is np.ndarray, \
            return an ak.Array when input is ak.Array. If "ak", return an ak.Array.

    Returns:
        parsed physical parameters. "x", "y", "z", "r" for position, \
            "pt", "px", "py", "pz", "p", "theta", "phi" for momentum, \
            "charge" for charge, "r_trk" for track radius.
    """
    helix0 = helix[..., 0]
    helix1 = helix[..., 1]
    helix2 = helix[..., 2]
    helix3 = helix[..., 3]
    helix4 = helix[..., 4]

    x = _helix01_to_x(helix0, helix1)
    y = _helix01_to_y(helix0, helix1)
    z = helix3
    r = np.abs(helix0)

    pt = _helix2_to_pt(helix2)
    px = _pt_helix1_to_px(pt, helix1)
    py = _pt_helix1_to_py(pt, helix1)
    pz = pt * helix4
    p = _pt_helix4_to_p(pt, helix4)
    theta = _pz_p_to_theta(pz, p)
    phi = np.atan2(py, px)

    charge = _helix2_to_charge(helix2)

    r_trk = pt * (10 / 2.99792458)  # |pt| * [GeV/c] / 1[e] / 1[T] = |pt| * 10/3 [m]
    r_trk = r_trk * 100  # to [cm]

    res_dict = {
        "x": x,
        "y": y,
        "z": z,
        "r": r,
        "px": px,
        "py": py,
        "pz": pz,
        "pt": pt,
        "p": p,
        "theta": theta,
        "phi": phi,
        "charge": charge,
        "r_trk": r_trk,
    }

    if isinstance(helix, ak.Array) or library == "ak":
        return ak.zip(res_dict)
    else:
        return res_dict



import numpy as np

def read_bes_tobject(
    data: np.ndarray,
    byte_offsets: np.ndarray,
    type_name: str,
    member_streamers_info: list[dict],
) -> list: ...
def read_bes_tobjarray(
    data: np.ndarray,
    byte_offsets: np.ndarray,
    type_name: str,
    member_streamers_info: list[dict],
) -> dict: ...
def read_bes_stl(
    data: np.ndarray,
    byte_offsets: np.ndarray,
    type_name: str,
    member_streamers_info: list[dict],
) -> dict: ...
def read_bes_raw(data: np.ndarray, sub_detectors: list[str]) -> dict: ...



from pathlib import Path
from typing import Literal

import awkward as ak
import numba as nb
import numpy as np

from ...typing import FloatLike, IntLike

_cur_dir = Path(__file__).resolve().parent

ENDCAP_PHI_01 = 64
ENDCAP_PHI_23 = 80
ENDCAP_PHI_45 = 96
ENDCAP_CRYSTALS = 480

BARREL_PHI = 120
BARREL_CRYSTALS = 5280


_emc_geom = dict(np.load(_cur_dir / "emc_geom.npz"))
_part: np.ndarray = _emc_geom["part"]
_theta: np.ndarray = _emc_geom["theta"]
_phi: np.ndarray = _emc_geom["phi"]
_points_x: np.ndarray = _emc_geom["points_x"]
_points_y: np.ndarray = _emc_geom["points_y"]
_points_z: np.ndarray = _emc_geom["points_z"]
_center_x: np.ndarray = _emc_geom["center_x"]
_center_y: np.ndarray = _emc_geom["center_y"]
_center_z: np.ndarray = _emc_geom["center_z"]
_front_center_x: np.ndarray = _emc_geom["front_center_x"]
_front_center_y: np.ndarray = _emc_geom["front_center_y"]
_front_center_z: np.ndarray = _emc_geom["front_center_z"]

emc_barrel_r = 94.2
emc_barrel_offset_1 = 2.5
emc_barrel_offset_2 = 5.0
emc_barrel_h1 = 5.1
emc_barrel_h2 = 5.2
emc_barrel_h3 = 5.2466
emc_barrel_l = 28.0


def get_emc_crystal_position(library: Literal["np", "ak", "pd"] = "np"):
    """
    Get EMC crystal position table.

    Parameters:
        library: The library to return the data in. Choose from 'ak', 'np', 'pd'.

    Returns:
        (ak.Array | dict[str, np.ndarray] | pd.DataFrame): The EMC crystal position table.

    Raises:
        ValueError: If the library is not 'ak', 'np', or 'pd'.
        ImportError: If the library is 'pd' but pandas is not installed.
    """
    cp: dict[str, np.ndarray] = {k: v.copy() for k, v in _emc_geom.items()}

    res: dict[str, np.ndarray] = {}

    for k in [
        "gid",
        "center_x",
        "center_y",
        "center_z",
        "front_center_x",
        "front_center_y",
        "front_center_z",
    ]:
        res[k] = cp[k]

    # flatten crystal points
    for i in range(8):
        res[f"points_x_{i}"] = cp["points_x"][:, i]
        res[f"points_y_{i}"] = cp["points_y"][:, i]
        res[f"points_z_{i}"] = cp["points_z"][:, i]

    if library == "ak":
        return ak.Array(res)
    elif library == "np":
        return res
    elif library == "pd":
        try:
            import pandas as pd  # type: ignore
        except ImportError:
            raise ImportError("Pandas is not installed. Run `pip install pandas`.")
        return pd.DataFrame(res)
    else:
        raise ValueError(f"Invalid library {library}. Choose from 'ak', 'np', 'pd'.")


@nb.vectorize([nb.uint16(nb.int_, nb.int_, nb.int_)])
def get_emc_gid(part: IntLike, theta: IntLike, phi: IntLike) -> IntLike:
    """
    Get EMC gid of given part, theta, and phi.

    - part 0: 0-479
        - theta 0: 0-63
        - theta 1: 64-127
        - theta 2: 128-207
        - theta 3: 208-287
        - theta 4: 288-383
        - theta 5: 384-479
    - part 1: 480-5759 (theta 0-47)
    - part 2: 5760-6239
        - theta 5: 5760-5855 (96)
        - theta 4: 5856-5951 (96)
        - theta 3: 5952-6031 (80)
        - theta 2: 6032-6111 (80)
        - theta 1: 6112-6175 (64)
        - theta 0: 6176-6239 (64)

    Parameters:
        part: part number
        theta: theta number
        phi: phi number

    Returns:
        index: EMC index
    """
    if part == 0:
        res = 0
        if theta == 0 or theta == 1:
            return theta * ENDCAP_PHI_01 + phi

        res += 2 * ENDCAP_PHI_01
        if theta == 2 or theta == 3:
            return res + (theta - 2) * ENDCAP_PHI_23 + phi

        res += 2 * ENDCAP_PHI_23
        if theta == 4 or theta == 5:
            return res + (theta - 4) * ENDCAP_PHI_45 + phi

    if part == 1:
        return ENDCAP_CRYSTALS + theta * BARREL_PHI + phi

    if part == 2:
        res = ENDCAP_CRYSTALS + BARREL_CRYSTALS

        if theta == 4 or theta == 5:
            return res + (5 - theta) * ENDCAP_PHI_45 + phi

        res += 2 * ENDCAP_PHI_45
        if theta == 2 or theta == 3:
            return res + (3 - theta) * ENDCAP_PHI_23 + phi

        res += 2 * ENDCAP_PHI_23
        if theta == 0 or theta == 1:
            return res + (1 - theta) * ENDCAP_PHI_01 + phi

    return 65535


@nb.vectorize([nb.uint8(nb.int_)])
def emc_gid_to_part(gid: IntLike) -> IntLike:
    """
    Convert EMC gid to part.

    Parameters:
        gid: The gid of the crystal.

    Returns:
        The part number of the crystal.
    """
    return _part[gid]


@nb.vectorize([nb.uint8(nb.int_)])
def emc_gid_to_theta(gid: IntLike) -> IntLike:
    """
    Convert EMC gid to theta.

    Parameters:
        gid: The gid of the crystal.

    Returns:
        The theta number of the crystal.
    """
    return _theta[gid]


@nb.vectorize([nb.uint8(nb.int_)])
def emc_gid_to_phi(gid: IntLike) -> IntLike:
    """
    Convert EMC gid to phi.

    Parameters:
        gid: The gid of the crystal.

    Returns:
        The phi number of the crystal.
    """
    return _phi[gid]


@nb.vectorize([nb.float64(nb.int_, nb.int_)])
def emc_gid_to_point_x(gid: IntLike, point: IntLike) -> FloatLike:
    """
    Convert EMC gid to x coordinate of the point.

    Parameters:
        gid: The gid of the crystal.
        point: The point number, 0-7.

    Returns:
        The x coordinate of the point.
    """
    return _points_x[gid, point]


@nb.vectorize([nb.float64(nb.int_, nb.int_)])
def emc_gid_to_point_y(gid: IntLike, point: IntLike) -> FloatLike:
    """
    Convert EMC gid to y coordinate of the point.

    Parameters:
        gid: The gid of the crystal.
        point: The point number, 0-7.

    Returns:
        The y coordinate of the point.
    """
    return _points_y[gid, point]


@nb.vectorize([nb.float64(nb.int_, nb.int_)])
def emc_gid_to_point_z(gid: IntLike, point: IntLike) -> FloatLike:
    """
    Convert EMC gid to z coordinate of the point.

    Parameters:
        gid: The gid of the crystal.
        point: The point number, 0-7.

    Returns:
        The z coordinate of the point.
    """
    return _points_z[gid, point]


@nb.vectorize([nb.float64(nb.int_)])
def emc_gid_to_center_x(gid: IntLike) -> FloatLike:
    """
    Convert EMC gid to x coordinate of the crystal's center.

    Parameters:
        gid: The gid of the crystal.

    Returns:
        The x coordinate of the crystal's center.
    """
    return _center_x[gid]


@nb.vectorize([nb.float64(nb.int_)])
def emc_gid_to_center_y(gid: IntLike) -> FloatLike:
    """
    Convert EMC gid to y coordinate of the crystal's center.

    Parameters:
        gid: The gid of the crystal.

    Returns:
        The y coordinate of the crystal's center.
    """
    return _center_y[gid]


@nb.vectorize([nb.float64(nb.int_)])
def emc_gid_to_center_z(gid: IntLike) -> FloatLike:
    """
    Convert EMC gid to z coordinate of the crystal's center.

    Parameters:
        gid: The gid of the crystal.

    Returns:
        The z coordinate of the crystal's center.
    """
    return _center_z[gid]


@nb.vectorize([nb.float64(nb.int_)])
def emc_gid_to_front_center_x(gid: IntLike) -> FloatLike:
    """
    Convert EMC gid to x coordinate of the crystal's front center.

    Parameters:
        gid: The gid of the crystal.

    Returns:
        The x coordinate of the crystal's front center.
    """
    return _front_center_x[gid]


@nb.vectorize([nb.float64(nb.int_)])
def emc_gid_to_front_center_y(gid: IntLike) -> FloatLike:
    """
    Convert EMC gid to y coordinate of the crystal's front center.

    Parameters:
        gid: The gid of the crystal.

    Returns:
        The y coordinate of the crystal's front center.
    """
    return _front_center_y[gid]


@nb.vectorize([nb.float64(nb.int_)])
def emc_gid_to_front_center_z(gid: IntLike) -> FloatLike:
    """
    Convert EMC gid to z coordinate of the crystal's front center.

    Parameters:
        gid: The gid of the crystal.

    Returns:
        The z coordinate of the crystal's front center.
    """
    return _front_center_z[gid]



from .emc import (
    emc_barrel_h1,
    emc_barrel_h2,
    emc_barrel_h3,
    emc_barrel_l,
    emc_barrel_offset_1,
    emc_barrel_offset_2,
    emc_barrel_r,
    emc_gid_to_center_x,
    emc_gid_to_center_y,
    emc_gid_to_center_z,
    emc_gid_to_front_center_x,
    emc_gid_to_front_center_y,
    emc_gid_to_front_center_z,
    emc_gid_to_part,
    emc_gid_to_phi,
    emc_gid_to_point_x,
    emc_gid_to_point_y,
    emc_gid_to_point_z,
    emc_gid_to_theta,
    get_emc_crystal_position,
    get_emc_gid,
)
from .mdc import (
    get_mdc_gid,
    get_mdc_wire_position,
    mdc_gid_to_east_x,
    mdc_gid_to_east_y,
    mdc_gid_to_east_z,
    mdc_gid_to_is_stereo,
    mdc_gid_to_layer,
    mdc_gid_to_stereo,
    mdc_gid_to_superlayer,
    mdc_gid_to_west_x,
    mdc_gid_to_west_y,
    mdc_gid_to_west_z,
    mdc_gid_to_wire,
    mdc_gid_z_to_x,
    mdc_gid_z_to_y,
    mdc_layer_to_is_stereo,
    mdc_layer_to_superlayer,
)
